from .discord_types.member import Member
from .discord_types.user import User
from .incoming import (
    IncomingDiscordInteraction,
    IncomingDiscordOption,
    IncomingDiscordOptionList,
    IncomingApplicationCommand,
)



from pydantic import BaseModel, ValidationError, validator
import typing
from .discord_types.member import Member

from ..register.models import CommandOption, SubcommandOption

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class IncomingDiscordOption(BaseModel):

    """An incoming discord option, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor will be accepted as a value in DiscordCommand
    """

    name: str
    value: str


class SubcommandIncomingDiscordOptionListChild(BaseModel):
    class Config:
        arbitary_types_allowed = True

    name: str
    options: typing.List[IncomingDiscordOption]


class SubcommandIncomingDiscordOptionList(BaseModel):
    """An incoming discord option list, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor accepted in DiscordCommand
    """

    class Config:
        arbitary_types_allowed = True

    name: str
    options: typing.Union[
        typing.List[IncomingDiscordOption],
        typing.List[SubcommandIncomingDiscordOptionListChild],
    ]


class IncomingDiscordOptionList(BaseModel):

    """An incoming discord option list, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor accepted in DiscordCommand
    """

    class Config:
        arbitary_types_allowed = True

    id: str
    name: str
    options: typing.Optional[
        typing.Union[
            typing.List[IncomingDiscordOption],
            typing.List[SubcommandIncomingDiscordOptionList],
        ]
    ] = None


class IncomingDiscordInteraction(BaseModel):

    """An incoming discord interaction, this is not intended for you to edit, and will not
    be accepted as an argument in any function.
    """

    class Config:
        arbitary_types_allowed = True

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordOptionList
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None


class IncomingApplicationCommand(BaseModel):

    """an Incoming Application command, this is not intended for you to edit, and will not be accepted
    in any function
    """

    class Config:
        arbitary_types_allowed = True

    id: int
    application_id: int
    name: str
    description: str
    options: typing.List[typing.Union[CommandOption, SubcommandOption]]


from httpx import Client
from .models import DiscordCommand
from loguru import logger
from ..errors.network import DiscordAPIError


class RegisterCommands(object):

    """This object contains methods to help registering a command to Discord.
    While you shouldn't need to import this directly, it's still accessible if you
    prefer not to initalize a Dispike object.

    Important to remember all methods are not async.
    """

    def __init__(self, application_id: str, bot_token: str):
        """Initalize object provided with application_id and a bot token

        Args:
            application_id (str): Client ID
            bot_token (str): Bot user Token
        """
        self.__bot_token = bot_token
        self._application_id = application_id
        self._client = Client(
            base_url=f"https://discord.com/api/v8/applications/{self._application_id}/"
        )

    def register(
        self, command: DiscordCommand, guild_only=False, guild_to_target: int = None
    ):
        """Register a completed `DiscordCommand` model to Discord API.

        Args:
            command (DiscordCommand): A properly configured DiscordCommand
            guild_only (bool, optional): Default to set global mode (True). Set to False to let the function know to expect a guild_id
            guild_to_target (int, optional): A guild Id if guild_only is set to True.
        """
        if guild_only == True:
            if guild_to_target is None:
                raise TypeError(
                    "if guild_only is set to true, a guild id must be provided."
                )

            logger.info(f"Targeting a specific guild -> {guild_to_target}")
            _request_url = f"guilds/{guild_to_target}/commands"
        else:
            _request_url = f"commands"

        try:
            _command_to_json = command.dict(exclude_none=True)
            _send_request = self._client.post(
                _request_url, headers=self.request_headers, json=_command_to_json
            )
            if _send_request.status_code in [200, 201]:
                return True

            raise DiscordAPIError(_send_request.status_code, _send_request.text)
        except Exception:
            raise

    @property
    def request_headers(self):
        """Return a valid header for authorization

        Returns:
            dict: a valid header for authorization
        """
        return {"Authorization": f"Bot {self.__bot_token}"}

    @property
    def bot_token(self):
        """You cannot view the bot_token directly, but you can still 'update' it.

        Raises:
            PermissionError: If you attempt to view the bot token without a new value
        """
        raise PermissionError("You cannot view the bot token directly.")

    @bot_token.setter
    def bot_token(self, new_bot_token):
        if new_bot_token == self.__bot_token:
            raise TypeError("Bot token already set to that.")
        self.__bot_token = new_bot_token


from .helper.embed import Embed
import typing
from .errors.network import DiscordAPIError
from loguru import logger

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal

if typing.TYPE_CHECKING:
    from .main import Dispike
    from .models import IncomingDiscordInteraction


class DiscordResponse(object):

    """Represents an outgoing Discord Response

    Attributes:
        content (str): A plain-text response to a user
        tts (bool): bool returning if the message should be spoken via tts.
        embeds (dict): a List representing .to_dict of an Embed object.
        response (dict): a valid response represented in a dict, to later be converted to JSON.
    """

    def __init__(
        self,
        content: str = None,
        tts: bool = False,
        embeds: typing.List[Embed] = [],
        show_user_input: bool = False,
        follow_up_message=False,
        empherical=False,
    ):
        """Initialize a DiscordResponse, you can either pass data into here, or
        simply create a DiscordResponse() and edit via properties.

        Args:
            content (str, optional): A plain-text response to a user
            tts (bool, optional): bool returning if the message should be spoken via tts
            embeds (typing.List[Embed], optional): a List representing .to_dict of an Embed object
            show_user_input (bool, optional): Whether to delete the user's message of calling the command after responding.
            empherical (bool, optional): Whether to send message as an empherical message.
        """
        if content != None:
            if isinstance(content, str) == False:
                raise TypeError(f"Content must be a string")
            elif content == "":
                content = None

        # if isinstance(content, str) == False or content == "" or content != None:
        #    raise TypeError(f"content must be a string. recieved: {content}")

        if isinstance(tts, bool) == False:
            raise TypeError("tts must be a bool")

        self._content = content
        self._tts = tts
        self._embeds = [x.to_dict() for x in embeds]
        if show_user_input == False:
            self._type_response = 3
        else:
            self._type_response = 4

        self._is_followup = follow_up_message
        self._is_empherical = empherical

    @property
    def embeds(self) -> typing.List[dict]:
        """Returns a list of embeds to send to.

        Returns:
            typing.List[dict]: Embeds represented as a dict.
        """
        # TODO: if accessing .embeds, return an Embed object instead of dict.
        return self._embeds

    def add_new_embed(self, embed_to_add: Embed):
        """Append a new embed, provided with a proper Embed object

        Args:
            embed_to_add (Embed): Proper Embed Object

        Raises:
            TypeError: Raised if you do not pass a proper Embed object.
        """
        if isinstance(embed_to_add, Embed):
            self._embeds.append(embed_to_add.to_dict())
        else:
            raise TypeError("embed must be a Embed object.")

    @property
    def content(self) -> str:
        """Either set or view the plain-text response to the user.

        Returns:
            str: Content provided
        """
        return self._content

    @content.setter
    def content(self, new_content_string: str):
        self._content = new_content_string

    @property
    def tts(self) -> bool:
        """Either set or view the tts attribute for the user.

        Returns:
            bool: tts
        """
        return self._tts

    @tts.setter
    def tts(self, new_tts: bool):
        self._tts = new_tts

    @property
    def response(self) -> dict:
        """A generated valid discord response

        Returns:
            dict: a valid discord response.
        """

        self.content = "" if self.content == None else self.content

        if self._is_followup:
            _req = {
                "content": self.content,
            }

            if self.embeds != []:
                _req["embeds"] = self.embeds

            if self.tts == True:
                _req["tts"] = True

            if self._is_empherical == True:
                logger.info("setting empherical")
                _req["flags"] == 1 << 6

            return _req

        _req = {
            "type": self._type_response,
            "data": {"tts": self.tts, "content": self.content, "embeds": self.embeds},
        }
        if self._is_empherical == True:
            _req["data"]["flags"] = 1 << 6
        return _req

    def _switch_to_followup_message(self):
        self._is_followup = True

    def __call__(self) -> dict:
        return self.response



import typing

# TODO: Automate stub generation until upstream issue fix.

class CommandTypes:
    SUB_COMMAND: int
    SUB_COMMAND_GROUP: int
    STRING: int
    INTEGER: int
    BOOLEAN: int
    USER: int
    CHANNEL: int
    ROLE: int

class CommandChoice:
    def __init__(self, name: str, value: str) -> None: ...

class CommandOption:
    def __init__(
        self,
        name: str,
        description: str,
        type: int,
        choices: typing.List[CommandChoice],
        required: bool = False,
    ) -> None: ...

class SubcommandOption:
    def __init__(
        self,
        name: str,
        description: str,
        options: typing.List[CommandOption],
        type: typing.Literal[2] = 2,
    ) -> None: ...

class DiscordCommand:
    def __init__(
        self,
        name: str,
        description: str,
        options: typing.List[typing.Union[SubcommandOption, CommandOption]],
    ) -> None: ...



from loguru import logger
from fastapi.responses import JSONResponse, Response
from fastapi import HTTPException
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

import typing

if typing.TYPE_CHECKING:
    from fastapi import FastAPI  # pragma: no cover


class DiscordVerificationMiddleware(BaseHTTPMiddleware):

    """Main middleware for verifying requests are signed by Discord.
    Per documentation.

    You should not need to import this directly.
    """

    def __init__(self, app: "FastAPI", *, client_public_key: str):
        """Initialize middleware

        Args:
            app (FastAPI): A valid, initialized FastAPI object
            client_public_key (str): a valid client public key
        """
        super().__init__(app)
        self._client_public_key = client_public_key
        logger.info(f"pub: {self._client_public_key}")
        self._verification_key = VerifyKey(bytes.fromhex(self._client_public_key))

    def verify_request(self, passed_signature: str, timestamp: str, body):
        """Verifies keys.

        Args:
            passed_signature (str): signature provided by discord in headers
            timestamp (str): timestamp provided by discord in headers
            body (TYPE): body provided by discord in headers

        Returns:
            tuple: bool,  status_code
        """
        try:
            message = timestamp.encode() + body
            self._verification_key.verify(message, bytes.fromhex(passed_signature))
            return True, 200
        except BadSignatureError:
            logger.error("bad signature")
            return False, 401
        except Exception:
            logger.exception("exception on verifying request")
            return False, 500

    async def dispatch(
        self,
        request: Request,
        call_next: typing.Callable[[Request], typing.Awaitable[Response]],
    ) -> Response:
        """Intercepts, verifies and dispatches request.

        Args:
            request (Request): Request object
            call_next (typing.Callable[[Request], typing.Awaitable[Response]]): next API endpoint
        """
        logger.debug("intercepting request.")

        if request.url.path == "/ping":
            logger.info("ping, forwarding")
            return await call_next(request)

        try:
            get_signature = request.headers["X-Signature-Ed25519"]
            get_timestamp = request.headers["X-Signature-Timestamp"]

            # https://github.com/encode/starlette/issues/495
            # this is bugged. calling request body will cause a timeout when the next endpoint tries to .body too
            get_body = await request.body()

            # so we need to store the body in the request state.
            request.state._cached_body = get_body
        except Exception:
            # logger.exception("error getting needed data for verification")
            return JSONResponse(
                status_code=400, content={"error_message": "Incorrect request."}
            )

        _status_bool, _status_code = self.verify_request(
            passed_signature=get_signature, timestamp=get_timestamp, body=get_body
        )
        if _status_bool == True:
            logger.info("approved request. forwarding call")
            _dispatch_request = await call_next(request)
            return _dispatch_request

        return JSONResponse(status_code=_status_code)



import typing

import httpx
from loguru import logger

from ..response import DiscordResponse
from ..errors.network import DiscordAPIError

if typing.TYPE_CHECKING:
    from ..main import Dispike  # pragma: no cover
    from ..models import IncomingDiscordInteraction  # pragma: no cover


class FollowUpMessages(object):
    def __init__(
        self,
        bot: "Dispike",
        interaction: "IncomingDiscordInteraction",
    ):
        """A module to handle Follow Up Messages.

        Args:
            bot (Dispike): An already initalized dispike object.
            interaction (IncomingDiscordInteraction): An incoming Discord Interaction
        """
        self._application_id = bot._application_id
        self._interaction_token = interaction.token

        self.base_url = f"https://discord.com/api/v8/webhooks/{self._application_id}/{self._interaction_token}"
        self._async_client = httpx.AsyncClient(base_url=self.base_url)
        self._sync_client = httpx.Client(base_url=self.base_url)

        self._message_id = None

    def create_follow_up_message(self, message: DiscordResponse):
        """Create an initial follow up message. (Sync)

        Args:
            message (DiscordResponse): An already created discord response

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if isinstance(message, DiscordResponse) == False:
            raise TypeError("Message must be a DiscordResponse")

        message._switch_to_followup_message()
        if self._message_id != None:
            raise TypeError("Creating a followup message can only be done once.")
        try:
            _request = self._sync_client.post(url=self.base_url, json=message.response)
            logger.info("sent request for creation of follow up to discord..")
            if _request.status_code in [200, 201]:
                _parse_request = _request.json()
                self._message_id = _parse_request["id"]
                return True
            else:
                logger.error(
                    f"discord returned a bad status code: {_request.status_code} -> {_request.text} url: {_request.url}"
                )
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise

    async def async_create_follow_up_message(self, message: DiscordResponse):
        """Create an initial follow up message. (Async)

        Args:
            message (DiscordResponse): An already created discord response

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if isinstance(message, DiscordResponse) == False:
            raise TypeError("Message must be a DiscordResponse")

        message._switch_to_followup_message()
        if self._message_id != None:
            raise TypeError("Creating a followup message can only be done once.")
        try:
            _request = await self._async_client.post(
                url=self.base_url, json=message.response
            )
            logger.info("sent request for creation of follow up to discord..")
            if _request.status_code in [200, 201]:
                _parse_request = _request.json()
                self._message_id = _parse_request["id"]
                return True
            else:
                logger.error(
                    f"discord returned a bad status code: {_request.status_code} -> {_request.text} url: {_request.url}"
                )
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise

    def edit_follow_up_message(self, updated_message: DiscordResponse):
        """Edit an already sent initial follow-up message. (Sync)

        Args:
            updated_message (DiscordResponse): An already created discord response

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if self._message_id == None:
            raise TypeError("a followup message must be sent first!")
        try:
            _request = self._sync_client.patch(
                f"/messages/{self._message_id}", json=updated_message.response
            )
            logger.info(
                f"sent request for edit of follow up to discord [{self._message_id}].."
            )
            if _request.status_code in [200, 201]:
                _parse_request = _request.json()
                self._message_id = _parse_request["id"]
                return True
            else:
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise

    async def async_edit_follow_up_message(self, updated_message: DiscordResponse):
        """Edit an already sent initial follow-up message. (Async)

        Args:
            updated_message (DiscordResponse): An already created discord response

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if self._message_id == None:
            raise TypeError("a followup message must be sent first!")
        try:
            _request = await self._async_client.patch(
                f"/messages/{self._message_id}", json=updated_message.response
            )
            logger.info(
                f"sent request for edit of follow up to discord [{self._message_id}].."
            )
            if _request.status_code in [200, 201]:
                _parse_request = _request.json()
                self._message_id = _parse_request["id"]
                return True
            else:
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise

    def delete_follow_up_message(self):
        """Deletes an already sent initial follow-up message. (sync)

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if self._message_id == None:
            raise TypeError("a followup message must be sent first.")
        try:
            _request = self._sync_client.delete(f"/messages/{self._message_id}")
            logger.info(
                f"sent request for deletion of follow up to discord [{self._message_id}].."
            )
            if _request.status_code in [200, 201, 204]:
                self._message_id = None
                return True
            else:
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise

    async def async_delete_follow_up_message(self):
        """Deletes an already sent initial follow-up message. (Async)

        Returns:
            True: If request is successfully made. An exception will rise otherwise

        Raises:
            DiscordAPIError: Discord returning a non-OK status status_code
            TypeError: Invalid type passed.
        """
        if self._message_id == None:
            raise TypeError("a followup message must be sent first.")
        try:
            _request = await self._async_client.delete(f"/messages/{self._message_id}")
            logger.info(
                f"sent request for deletion of follow up to discord [{self._message_id}].."
            )
            if _request.status_code in [200, 201, 204]:
                self._message_id = None
                return True
            else:
                raise DiscordAPIError(_request.status_code, _request.text)
        except DiscordAPIError:
            raise
        except Exception:
            logger.exception("error creating message.")
            raise